from __future__ import annotations

import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Any


def ensure_dir(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def write_json(path: Path, data: Any) -> None:
    ensure_dir(path.parent)
    path.write_text(json.dumps(data, indent=2, default=str) + "\n", encoding="utf-8")


def read_json(path: Path) -> Any:
    return json.loads(path.read_text(encoding="utf-8"))


def write_text(path: Path, text: str) -> None:
    ensure_dir(path.parent)
    path.write_text(text, encoding="utf-8")


def load_changed_files(path: Path) -> list[str]:
    content = path.read_text(encoding="utf-8").strip()
    if not content:
        return []
    if content.startswith("["):
        parsed = json.loads(content)
        if not isinstance(parsed, list):
            return []
        return [f for f in parsed if isinstance(f, str)]
    return [line.strip() for line in content.splitlines() if line.strip()]


def now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()



from __future__ import annotations

import re
import shlex
from pathlib import Path

from pygate.exec import run_command
from pygate.models import FailuresPayload, GateName

_EXCLUDED_DIRS = re.compile(r"(^|/)(\.(pygate|git|venv|tox|nox)|__pycache__|dist|build|node_modules)(/|$)")
_MAX_FILES = 20


def _is_eligible(path: str) -> bool:
    if not path.endswith(".py"):
        return False
    if _EXCLUDED_DIRS.search(path):
        return False
    return not (path.startswith("/") or ".." in path)


def _collect_scoped_files(failures: FailuresPayload) -> list[str]:
    seen: set[str] = set()
    files: list[str] = []

    for f in failures.changed_files:
        if f not in seen and _is_eligible(f):
            seen.add(f)
            files.append(f)

    for finding in failures.findings:
        for f in finding.files:
            if f not in seen and _is_eligible(f):
                seen.add(f)
                files.append(f)

    return files[:_MAX_FILES]


def run_deterministic_prefix(*, cwd: Path, failures: FailuresPayload) -> list[dict]:
    actions: list[dict] = []

    has_lint_failure = any(f.gate == GateName.LINT for f in failures.findings)
    if not has_lint_failure:
        return actions

    scoped_files = _collect_scoped_files(failures)
    if not scoped_files:
        return actions

    file_args = " ".join(shlex.quote(f) for f in scoped_files)

    # ruff check --fix (safe fixes only)
    fix_trace = run_command(f"ruff check --fix {file_args}", cwd=cwd)
    actions.append(
        {
            "rule_id": "RUFF_AUTOFIX",
            "strategy": "deterministic_prefix",
            "accepted": fix_trace.exit_code in (0, 1),  # ruff returns 1 if unfixed issues remain
            "command": fix_trace.command,
            "exit_code": fix_trace.exit_code,
            "files": scoped_files,
            "rationale": "Apply safe ruff fixes on scoped files to clear auto-fixable lint issues.",
        }
    )

    # ruff format
    fmt_trace = run_command(f"ruff format {file_args}", cwd=cwd)
    actions.append(
        {
            "rule_id": "RUFF_FORMAT",
            "strategy": "deterministic_prefix",
            "accepted": fmt_trace.exit_code == 0,
            "command": fmt_trace.command,
            "exit_code": fmt_trace.exit_code,
            "files": scoped_files,
            "rationale": "Apply ruff formatting on scoped files.",
        }
    )

    return actions



from __future__ import annotations

import shutil
import time
from pathlib import Path

from pygate.config import load_config
from pygate.constants import (
    ESCALATION_FILE,
    FAILURES_FILE,
    PYGATE_DIR,
    REPAIR_REPORT_FILE,
    EscalationCode,
)
from pygate.deterministic_fix import run_deterministic_prefix
from pygate.env import command_exists
from pygate.exec import run_command
from pygate.fs import ensure_dir, read_json, write_json
from pygate.models import (
    Escalation,
    FailuresPayload,
    RepairAttempt,
    RepairReport,
)
from pygate.run_command import execute_run
from pygate.summarize_command import execute_summarize

_BACKUP_EXCLUDES = {".pygate", ".git", "__pycache__", ".venv", ".tox", ".nox", "dist", "node_modules"}


def _backup_workspace(cwd: Path, backup_dir: Path) -> None:
    if backup_dir.exists():
        shutil.rmtree(backup_dir)
    shutil.copytree(cwd, backup_dir, symlinks=True, ignore=shutil.ignore_patterns(*_BACKUP_EXCLUDES))


def _restore_workspace(cwd: Path, backup_dir: Path) -> None:
    for item in cwd.iterdir():
        if item.name in _BACKUP_EXCLUDES:
            continue
        if item.is_dir():
            shutil.rmtree(item)
        else:
            item.unlink()
    for item in backup_dir.iterdir():
        dest = cwd / item.name
        if item.is_dir():
            shutil.copytree(item, dest)
        else:
            shutil.copy2(item, dest)


def _diff_snapshot(cwd: Path) -> dict[str, int]:
    if not command_exists("git"):
        return {}
    trace = run_command(
        "git diff --numstat -- . ':(exclude).pygate' ':(exclude)__pycache__' ':(exclude).venv'",
        cwd=cwd,
    )
    if trace.exit_code != 0:
        return {}
    result: dict[str, int] = {}
    for line in trace.stdout.strip().splitlines():
        parts = line.split("\t")
        if len(parts) == 3:
            added = int(parts[0]) if parts[0] != "-" else 0
            removed = int(parts[1]) if parts[1] != "-" else 0
            result[parts[2]] = added + removed
    return result


def _compute_patch_lines(before: dict[str, int], after: dict[str, int]) -> int:
    all_files = set(before) | set(after)
    total = 0
    for f in all_files:
        total += abs(after.get(f, 0) - before.get(f, 0))
    return total


def _escalate(reason_code: str, message: str, evidence: dict, cwd: Path) -> dict:
    escalation = Escalation(
        reason_code=reason_code,
        message=message,
        evidence=evidence,
    )
    write_json(cwd / ESCALATION_FILE, escalation.model_dump(mode="json"))
    return escalation.model_dump(mode="json")


def execute_repair(
    *,
    input_path: str,
    max_attempts: int | None = None,
    cwd: Path | None = None,
) -> dict:
    cwd = cwd or Path.cwd()
    pygate_dir = cwd / PYGATE_DIR
    ensure_dir(pygate_dir)

    config = load_config(cwd)
    policy = config.get("policy", {})
    max_attempts = max_attempts if max_attempts is not None else policy.get("max_attempts", 3)
    max_patch_lines = policy.get("max_patch_lines", 150)
    abort_threshold = policy.get("abort_on_no_improvement", 2)
    time_cap_seconds = policy.get("time_cap_seconds", 1200)

    failures_data = read_json(Path(input_path))
    failures = FailuresPayload(**failures_data)

    previous_count = len(failures.findings)
    no_improvement = 0
    attempts: list[RepairAttempt] = []
    started = time.monotonic()

    # Detect mode from failures for re-runs
    mode = failures.mode

    for attempt_num in range(1, max_attempts + 1):
        # Time cap check
        elapsed = time.monotonic() - started
        if elapsed > time_cap_seconds:
            return _escalate(
                EscalationCode.UNKNOWN_BLOCKER,
                f"Time cap reached ({time_cap_seconds}s).",
                {"elapsed_seconds": int(elapsed)},
                cwd,
            )

        # Backup
        backup_dir = pygate_dir / f"backup-attempt-{attempt_num}"
        _backup_workspace(cwd, backup_dir)

        # Snapshot before
        before_snap = _diff_snapshot(cwd)

        # Run deterministic fixes
        actions = run_deterministic_prefix(cwd=cwd, failures=failures)

        # If no fix strategies applied, don't waste attempts
        if not actions and attempt_num > 1:
            return _escalate(
                EscalationCode.NO_IMPROVEMENT,
                f"No applicable fix strategies for remaining {previous_count} finding(s).",
                {
                    "attempts": [a.model_dump(mode="json") for a in attempts],
                    "latest_failures_path": str(cwd / FAILURES_FILE),
                    "remaining_gates": list({f.gate.value for f in failures.findings}),
                },
                cwd,
            )

        # Snapshot after
        after_snap = _diff_snapshot(cwd)
        patch_lines = _compute_patch_lines(before_snap, after_snap)

        # Patch budget check
        if patch_lines > max_patch_lines:
            _restore_workspace(cwd, backup_dir)
            return _escalate(
                EscalationCode.PATCH_BUDGET_EXCEEDED,
                f"Patch exceeded budget: {patch_lines} lines > {max_patch_lines} max.",
                {
                    "attempt": attempt_num,
                    "patch_lines": patch_lines,
                    "max_patch_lines": max_patch_lines,
                },
                cwd,
            )

        # Re-run gates and regenerate agent brief
        rerun = execute_run(mode=mode, changed_files=failures.changed_files, cwd=cwd)
        execute_summarize(input_path=rerun["failures_path"], cwd=cwd)
        rerun_failures_data = read_json(Path(rerun["failures_path"]))
        rerun_failures = FailuresPayload(**rerun_failures_data)
        current_count = len(rerun_failures.findings)

        improved = current_count < previous_count
        worsened = current_count > previous_count

        attempts.append(
            RepairAttempt(
                attempt=attempt_num,
                patch_lines=patch_lines,
                before_findings=previous_count,
                after_findings=current_count,
                improved=improved,
                worsened=worsened,
                status=rerun["status"],
                actions=list(actions),
            )
        )

        # Pass check
        if rerun["status"] == "pass":
            report = RepairReport(status="pass", attempts=attempts)
            write_json(cwd / REPAIR_REPORT_FILE, report.model_dump(mode="json"))
            return report.model_dump(mode="json")

        # Rollback if worsened
        if worsened:
            _restore_workspace(cwd, backup_dir)
            # Don't update previous_count or failures — workspace was restored
        else:
            previous_count = current_count
            failures = rerun_failures

        # Track improvement
        if improved:
            no_improvement = 0
        else:
            no_improvement += 1

        # No improvement abort
        if no_improvement >= abort_threshold:
            return _escalate(
                EscalationCode.NO_IMPROVEMENT,
                f"No measurable improvement for {no_improvement} consecutive attempt(s).",
                {
                    "attempts": [a.model_dump(mode="json") for a in attempts],
                    "latest_failures_path": str(cwd / FAILURES_FILE),
                },
                cwd,
            )

    # Attempts exhausted
    return _escalate(
        EscalationCode.UNKNOWN_BLOCKER,
        f"Attempts exhausted ({max_attempts}).",
        {"latest_failures_path": str(cwd / FAILURES_FILE)},
        cwd,
    )



from __future__ import annotations

import json
from pathlib import Path

from pygate.models import Finding, GateName, Severity

LONGREPR_MAX = 500


def resolve_pytest_command(commands_config: dict, cwd: Path) -> str:
    if "test" in commands_config:
        return commands_config["test"]
    report_path = cwd / ".pygate" / "pytest-report.json"
    import shlex

    return f"pytest --json-report --json-report-file={shlex.quote(str(report_path))} -q"


def parse_pytest_output(
    stdout: str,
    stderr: str,
    exit_code: int,
    report_path: Path | None,
    cwd: Path,
) -> list[Finding]:
    if report_path and report_path.exists():
        return _parse_json_report(report_path, exit_code, cwd)
    return []


def _parse_json_report(report_path: Path, exit_code: int, cwd: Path) -> list[Finding]:
    try:
        data = json.loads(report_path.read_text(encoding="utf-8"))
    except (json.JSONDecodeError, ValueError, OSError):
        return []

    if not isinstance(data, dict):
        return []

    tests = data.get("tests", [])
    findings: list[Finding] = []

    for test in tests:
        outcome = test.get("outcome", "")
        if outcome not in ("failed", "error"):
            continue

        nodeid = test.get("nodeid", "")
        call = test.get("call", {})
        longrepr = call.get("longrepr", "")
        duration = call.get("duration", 0)

        # Extract file path from nodeid (e.g., "tests/test_foo.py::test_bar")
        file_part = nodeid.split("::")[0] if "::" in nodeid else nodeid

        # Sanitize nodeid for use as finding id
        safe_id = nodeid.replace("::", "_").replace("/", "_").replace(".", "_")
        finding_id = f"pytest_{safe_id}"

        # Truncate longrepr
        truncated = longrepr[:LONGREPR_MAX] + "..." if len(longrepr) > LONGREPR_MAX else longrepr

        # Short summary from first line of longrepr or just the nodeid
        first_line = longrepr.split("\n")[0][:200] if longrepr else ""
        summary = f"{nodeid}: {first_line}" if first_line else f"{nodeid} failed"

        findings.append(
            Finding(
                id=finding_id,
                gate=GateName.TEST,
                severity=Severity.HIGH,
                summary=summary,
                files=[file_part] if file_part else [],
                actual="failed",
                threshold="passed",
                raw={
                    "longrepr": truncated,
                    "duration": duration,
                    "outcome": outcome,
                },
            )
        )

    return findings



from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

from pygate import __version__
from pygate.env import check_environment
from pygate.fs import load_changed_files
from pygate.models import RunMode
from pygate.repair_command import execute_repair
from pygate.run_command import execute_run
from pygate.summarize_command import execute_summarize


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="pygate",
        description="Deterministic quality gate CLI for Python projects",
    )
    parser.add_argument("--version", action="version", version=f"pygate {__version__}")

    sub = parser.add_subparsers(dest="command")

    # run
    run_p = sub.add_parser("run", help="Run quality gates")
    run_p.add_argument("--mode", required=True, choices=["canary", "full"], help="Gate mode")
    run_p.add_argument("--changed-files", required=True, help="Path to changed files list")

    # summarize
    sum_p = sub.add_parser("summarize", help="Generate agent brief from failures")
    sum_p.add_argument("--input", required=True, help="Path to failures.json")

    # repair
    rep_p = sub.add_parser("repair", help="Run bounded repair loop")
    rep_p.add_argument("--input", required=True, help="Path to failures.json")
    rep_p.add_argument("--max-attempts", type=int, default=None, help="Maximum repair attempts")

    return parser


def main(argv: list[str] | None = None) -> None:
    parser = _build_parser()
    args = parser.parse_args(argv)

    if not args.command:
        parser.print_help()
        sys.exit(0)

    # Print environment warnings
    warnings = check_environment(command=args.command)
    for w in warnings:
        print(f"[pygate warn] {w}", file=sys.stderr)

    cwd = Path.cwd()

    if args.command == "run":
        changed_files_path = Path(args.changed_files)
        if not changed_files_path.is_absolute():
            changed_files_path = cwd / changed_files_path
        changed_files = load_changed_files(changed_files_path)

        mode = RunMode(args.mode)
        result = execute_run(mode=mode, changed_files=changed_files, cwd=cwd)
        print(json.dumps(result, indent=2))
        sys.exit(0 if result["status"] == "pass" else 1)

    elif args.command == "summarize":
        input_path = args.input
        if not Path(input_path).is_absolute():
            input_path = str(cwd / input_path)
        result = execute_summarize(input_path=input_path, cwd=cwd)
        print(json.dumps(result, indent=2))
        sys.exit(0)

    elif args.command == "repair":
        input_path = args.input
        if not Path(input_path).is_absolute():
            input_path = str(cwd / input_path)
        result = execute_repair(
            input_path=input_path,
            max_attempts=args.max_attempts,
            cwd=cwd,
        )
        print(json.dumps(result, indent=2))
        status = result.get("status", "")
        if status == "pass":
            sys.exit(0)
        else:
            # execute_repair returns "pass" or "escalated"
            sys.exit(2)



from __future__ import annotations

PYGATE_DIR = ".pygate"
FAILURES_FILE = ".pygate/failures.json"
RUN_METADATA_FILE = ".pygate/run-metadata.json"
AGENT_BRIEF_JSON_FILE = ".pygate/agent-brief.json"
AGENT_BRIEF_MD_FILE = ".pygate/agent-brief.md"
REPAIR_REPORT_FILE = ".pygate/repair-report.json"
ESCALATION_FILE = ".pygate/escalation.json"
PYTEST_REPORT_FILE = ".pygate/pytest-report.json"

DEFAULT_POLICY: dict[str, int] = {
    "max_attempts": 3,
    "max_patch_lines": 150,
    "abort_on_no_improvement": 2,
    "time_cap_seconds": 1200,
}


class EscalationCode:
    NO_IMPROVEMENT = "NO_IMPROVEMENT"
    PATCH_BUDGET_EXCEEDED = "PATCH_BUDGET_EXCEEDED"
    UNKNOWN_BLOCKER = "UNKNOWN_BLOCKER"
    UNRESOLVED_DETERMINISTIC_FAILURES = "UNRESOLVED_DETERMINISTIC_FAILURES"
    # Reserved for future use (v2 model-assisted repair)
    ARCHITECTURAL_CHANGE_REQUIRED = "ARCHITECTURAL_CHANGE_REQUIRED"
    FLAKY_EVALUATOR = "FLAKY_EVALUATOR"
    ENVIRONMENT_DRIFT = "ENVIRONMENT_DRIFT"
    TEST_FIXTURE_OR_EXTERNAL_DEP = "TEST_FIXTURE_OR_EXTERNAL_DEP"



from __future__ import annotations

import json
from pathlib import Path

from pygate.models import Finding, GateName, Severity


def resolve_pyright_command(commands_config: dict) -> str:
    return commands_config.get("typecheck", "pyright --outputjson .")


def _map_severity(pyright_severity: str) -> Severity:
    match pyright_severity:
        case "error":
            return Severity.HIGH
        case "warning":
            return Severity.MEDIUM
        case "information":
            return Severity.LOW
        case _:
            return Severity.MEDIUM


def parse_pyright_output(stdout: str, stderr: str, exit_code: int, cwd: Path) -> list[Finding]:
    try:
        data = json.loads(stdout)
    except (json.JSONDecodeError, ValueError):
        return []

    if not isinstance(data, dict):
        return []

    diagnostics = data.get("generalDiagnostics", [])

    findings: list[Finding] = []
    for d in diagnostics:
        sev = d.get("severity", "error")
        if sev == "information":
            continue

        file_path = d.get("file", "")
        message = d.get("message", "")
        rule = d.get("rule", "")
        range_info = d.get("range", {})
        start = range_info.get("start", {})
        line = start.get("line", 0) + 1  # pyright is 0-indexed
        col = start.get("character", 0) + 1  # pyright is 0-indexed

        try:
            rel_path = str(Path(file_path).relative_to(cwd))
        except ValueError:
            rel_path = file_path

        finding_id = f"pyright_{rule}_{rel_path}_{line}" if rule else f"pyright_{rel_path}_{line}"

        findings.append(
            Finding(
                id=finding_id,
                gate=GateName.TYPECHECK,
                severity=_map_severity(sev),
                summary=f"{rule}: {message}" if rule else message,
                files=[rel_path],
                rule=rule or None,
                line=line,
                column=col,
                actual=1,
                threshold=0,
                raw={
                    "pyright_severity": sev,
                    "range": range_info,
                },
            )
        )

    return findings



from __future__ import annotations

import json
from pathlib import Path

from pygate.models import Finding, GateName, Severity


def resolve_ruff_command(commands_config: dict) -> str:
    return commands_config.get("lint", "ruff check --output-format json --exclude .pygate .")


def _severity_for_code(code: str) -> Severity:
    if not code:
        return Severity.MEDIUM
    prefix = code[0].upper()
    if prefix in ("E", "F"):
        return Severity.HIGH
    if prefix == "W":
        return Severity.MEDIUM
    return Severity.MEDIUM


def parse_ruff_output(stdout: str, stderr: str, exit_code: int, cwd: Path) -> list[Finding]:
    try:
        violations = json.loads(stdout)
    except (json.JSONDecodeError, ValueError):
        return []

    if not isinstance(violations, list):
        return []

    findings: list[Finding] = []
    for v in violations:
        code = v.get("code", "")
        filename = v.get("filename", "")
        location = v.get("location", {})
        row = location.get("row", 0)
        col = location.get("column", 0)
        message = v.get("message", "")
        fix_info = v.get("fix") or {}

        # Relativize filename
        try:
            rel_path = str(Path(filename).relative_to(cwd))
        except ValueError:
            rel_path = filename

        idx = len(findings)
        finding_id = f"ruff_{code}_{rel_path}_{row}_{idx}"

        findings.append(
            Finding(
                id=finding_id,
                gate=GateName.LINT,
                severity=_severity_for_code(code),
                summary=f"{code}: {message}",
                files=[rel_path],
                rule=code,
                line=row,
                column=col,
                actual=exit_code,
                threshold=0,
                raw={
                    "fix_applicability": fix_info.get("applicability"),
                    "url": v.get("url"),
                },
            )
        )

    return findings



from __future__ import annotations

import sys
from pathlib import Path

from pygate.constants import DEFAULT_POLICY

if sys.version_info >= (3, 11):
    import tomllib
else:
    try:
        import tomli as tomllib  # pyright: ignore[reportMissingImports]
    except ImportError:
        tomllib = None  # type: ignore[assignment]


def load_config(cwd: Path | None = None) -> dict:
    cwd = cwd or Path.cwd()

    # Try pygate.toml first
    pygate_toml = cwd / "pygate.toml"
    if pygate_toml.exists():
        return _merge_config(_load_toml(pygate_toml), source=str(pygate_toml))

    # Try pyproject.toml [tool.pygate]
    pyproject = cwd / "pyproject.toml"
    if pyproject.exists():
        full = _load_toml(pyproject)
        user_config = full.get("tool", {}).get("pygate", {})
        if user_config:
            return _merge_config(user_config, source=str(pyproject))

    return _defaults()


def _load_toml(path: Path) -> dict:
    if tomllib is None:
        raise ImportError(
            f"Cannot parse {path}: tomli package required on Python < 3.11. Install it with: pip install tomli"
        )
    with open(path, "rb") as f:
        return tomllib.load(f)


def _defaults() -> dict:
    return {
        "policy": dict(DEFAULT_POLICY),
        "commands": {},
        "gates": {},
        "source": "defaults",
    }


def _merge_config(user: dict, *, source: str) -> dict:
    defaults = _defaults()
    policy = {**defaults["policy"], **user.get("policy", {})}
    commands = user.get("commands", {})
    gates = {**defaults["gates"], **user.get("gates", {})}
    return {
        "policy": policy,
        "commands": commands,
        "gates": gates,
        "source": source,
    }



from __future__ import annotations

from pathlib import Path

from pygate.exec import run_command
from pygate.gates.pyright import parse_pyright_output, resolve_pyright_command
from pygate.gates.pytest_gate import parse_pytest_output, resolve_pytest_command
from pygate.gates.ruff import parse_ruff_output, resolve_ruff_command
from pygate.models import (
    CommandTrace,
    Finding,
    GateName,
    GateResult,
    GateStatus,
    RunMode,
    Severity,
)


def _finding_for_exit_code(gate: GateName, trace: CommandTrace) -> Finding:
    return Finding(
        id=f"{gate.value}_exit_{trace.exit_code}",
        gate=gate,
        severity=Severity.HIGH,
        summary=f"{gate.value} command failed with exit code {trace.exit_code}",
        actual=trace.exit_code,
        threshold=0,
        raw={
            "command": trace.command,
            "stderr_excerpt": "\n".join(trace.stderr.splitlines()[:30]),
            "stdout_excerpt": "\n".join(trace.stdout.splitlines()[:30]),
        },
    )


def run_deterministic_gates(
    *,
    mode: RunMode,
    cwd: Path,
    config: dict,
    changed_files: list[str],
) -> tuple[list[GateResult], list[Finding], list[CommandTrace]]:
    gates_config = config.get("gates", {})
    commands_config = config.get("commands", {})
    test_in_canary = gates_config.get("test_in_canary", False)

    gate_plan = [
        (GateName.LINT, True),
        (GateName.TYPECHECK, True),
        (GateName.TEST, mode == RunMode.FULL or test_in_canary),
    ]

    gate_results: list[GateResult] = []
    all_findings: list[Finding] = []
    all_traces: list[CommandTrace] = []

    for gate_name, enabled in gate_plan:
        if not enabled:
            gate_results.append(GateResult(name=gate_name, status=GateStatus.SKIPPED, duration_ms=0))
            continue

        cmd = _resolve_command(gate_name, commands_config, cwd)
        trace = run_command(cmd, cwd=cwd)
        all_traces.append(trace)

        if trace.exit_code != 0:
            findings = _parse_gate_output(gate_name, trace, cwd)
            if not findings:
                findings = [_finding_for_exit_code(gate_name, trace)]
            all_findings.extend(findings)
            gate_results.append(GateResult(name=gate_name, status=GateStatus.FAIL, duration_ms=trace.duration_ms))
        else:
            gate_results.append(GateResult(name=gate_name, status=GateStatus.PASS, duration_ms=trace.duration_ms))

    return gate_results, all_findings, all_traces


def _resolve_command(gate: GateName, commands_config: dict, cwd: Path) -> str:
    if gate.value in commands_config:
        return commands_config[gate.value]

    match gate:
        case GateName.LINT:
            return resolve_ruff_command(commands_config)
        case GateName.TYPECHECK:
            return resolve_pyright_command(commands_config)
        case GateName.TEST:
            return resolve_pytest_command(commands_config, cwd)
        case _:
            raise ValueError(f"Unknown gate: {gate}")


def _parse_gate_output(gate: GateName, trace: CommandTrace, cwd: Path) -> list[Finding]:
    match gate:
        case GateName.LINT:
            return parse_ruff_output(trace.stdout, trace.stderr, trace.exit_code, cwd)
        case GateName.TYPECHECK:
            return parse_pyright_output(trace.stdout, trace.stderr, trace.exit_code, cwd)
        case GateName.TEST:
            report_path = cwd / ".pygate" / "pytest-report.json"
            return parse_pytest_output(trace.stdout, trace.stderr, trace.exit_code, report_path, cwd)
        case _:
            return []



from __future__ import annotations

import os
import subprocess
import time
from pathlib import Path

from pygate.fs import now_iso
from pygate.models import CommandTrace


def run_command(
    command: str,
    *,
    cwd: Path | str | None = None,
    timeout_seconds: int | None = None,
    env: dict[str, str] | None = None,
) -> CommandTrace:
    work_dir = str(cwd) if cwd else os.getcwd()
    merged_env = {**os.environ, **(env or {})}
    started_at = now_iso()
    start = time.monotonic()
    timed_out = False

    try:
        result = subprocess.run(
            command,
            shell=True,
            cwd=work_dir,
            capture_output=True,
            text=True,
            timeout=timeout_seconds,
            env=merged_env,
        )
        exit_code = result.returncode
        stdout = result.stdout or ""
        stderr = result.stderr or ""
    except subprocess.TimeoutExpired as exc:
        timed_out = True
        exit_code = 1
        stdout = exc.stdout.decode("utf-8", errors="replace") if isinstance(exc.stdout, bytes) else (exc.stdout or "")
        stderr = exc.stderr.decode("utf-8", errors="replace") if isinstance(exc.stderr, bytes) else (exc.stderr or "")

    duration_ms = int((time.monotonic() - start) * 1000)

    return CommandTrace(
        command=command,
        cwd=work_dir,
        started_at=started_at,
        duration_ms=duration_ms,
        exit_code=exit_code,
        timed_out=timed_out,
        stdout=stdout,
        stderr=stderr,
    )



from __future__ import annotations

import json
import shutil
import sys

from pygate.exec import run_command
from pygate.models import EnvironmentInfo


def command_exists(name: str) -> bool:
    return shutil.which(name) is not None


def check_environment(*, command: str) -> list[str]:
    warnings: list[str] = []
    if not command_exists("git"):
        warnings.append("git not found: repo metadata will be unavailable")
    if command == "run":
        if not command_exists("ruff"):
            warnings.append("ruff not found: lint gate will fail")
        if not command_exists("pyright"):
            warnings.append("pyright not found: typecheck gate will fail")
    return warnings


def capture_environment() -> EnvironmentInfo:
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    platform = sys.platform
    virtualenv = sys.prefix if sys.prefix != sys.base_prefix else None
    resolver = _detect_resolver()
    installed = _get_installed_packages()

    return EnvironmentInfo(
        python_version=python_version,
        platform=platform,
        virtualenv=virtualenv,
        resolver=resolver,
        installed_packages=installed,
    )


def _detect_resolver() -> str | None:
    if command_exists("uv"):
        return "uv"
    if command_exists("poetry"):
        return "poetry"
    if command_exists("pip"):
        return "pip"
    return None


def _get_installed_packages() -> dict[str, str]:
    for cmd in ["uv pip list --format json", "pip list --format json"]:
        exe = cmd.split()[0]
        if not command_exists(exe):
            continue
        trace = run_command(cmd, timeout_seconds=30)
        if trace.exit_code == 0 and trace.stdout.strip():
            try:
                pkgs = json.loads(trace.stdout)
                return {p["name"]: p["version"] for p in pkgs if isinstance(p, dict)}
            except (json.JSONDecodeError, KeyError):
                continue
    return {}



from __future__ import annotations

import time
import uuid
from datetime import datetime, timezone
from pathlib import Path

from pygate.config import load_config
from pygate.constants import FAILURES_FILE, PYGATE_DIR, RUN_METADATA_FILE
from pygate.env import capture_environment
from pygate.exec import run_command
from pygate.fs import ensure_dir, now_iso, write_json
from pygate.gates import run_deterministic_gates
from pygate.models import (
    Confidence,
    FailuresPayload,
    InferredHint,
    RunMetadata,
    RunMode,
    RunStatus,
)


def _git_info(cwd: Path) -> dict[str, str | None]:
    from pygate.env import command_exists

    if not command_exists("git"):
        return {"repo": None, "branch": None}

    repo = None
    branch = None

    trace = run_command("git config --get remote.origin.url", cwd=cwd)
    if trace.exit_code == 0:
        repo = trace.stdout.strip() or None

    trace = run_command("git rev-parse --abbrev-ref HEAD", cwd=cwd)
    if trace.exit_code == 0:
        branch = trace.stdout.strip() or None

    return {"repo": repo, "branch": branch}


def _generate_run_id() -> str:
    ts = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    short_uuid = uuid.uuid4().hex[:8]
    return f"run_{ts}_{short_uuid}"


def execute_run(
    *,
    mode: RunMode,
    changed_files: list[str],
    cwd: Path | None = None,
) -> dict:
    cwd = cwd or Path.cwd()
    pygate_dir = cwd / PYGATE_DIR
    ensure_dir(pygate_dir)

    run_id = _generate_run_id()
    started_at = now_iso()
    start_ms = _monotonic_ms()

    config = load_config(cwd)
    environment = capture_environment()

    gate_results, findings, traces = run_deterministic_gates(
        mode=mode, cwd=cwd, config=config, changed_files=changed_files
    )

    status = RunStatus.FAIL if findings else RunStatus.PASS
    git = _git_info(cwd)

    inferred_hints = [
        InferredHint(
            finding_id=f.id,
            hint=f"Start with the deterministic gate failure in {f.gate.value}. "
            "Inspect command output in run-metadata traces.",
            confidence=Confidence.LOW,
        )
        for f in findings
    ]

    failures = FailuresPayload(
        run_id=run_id,
        mode=mode,
        status=status,
        timestamp=now_iso(),
        repo=git["repo"],
        branch=git["branch"],
        changed_files=changed_files,
        gates=gate_results,
        findings=findings,
        inferred_hints=inferred_hints,
    )

    metadata = RunMetadata(
        run_id=run_id,
        mode=mode,
        started_at=started_at,
        completed_at=now_iso(),
        duration_ms=_monotonic_ms() - start_ms,
        config_source=config.get("source", "defaults"),
        environment=environment,
        command_traces=traces,
    )

    failures_path = cwd / FAILURES_FILE
    metadata_path = cwd / RUN_METADATA_FILE

    write_json(failures_path, failures.model_dump(mode="json"))
    write_json(metadata_path, metadata.model_dump(mode="json"))

    return {
        "status": status.value,
        "failures_path": str(failures_path),
        "metadata_path": str(metadata_path),
        "run_id": run_id,
    }


def _monotonic_ms() -> int:
    return int(time.monotonic() * 1000)



from __future__ import annotations

from enum import Enum
from typing import Any

from pydantic import BaseModel, Field


class GateStatus(str, Enum):
    PASS = "pass"
    FAIL = "fail"
    SKIPPED = "skipped"


class RunMode(str, Enum):
    CANARY = "canary"
    FULL = "full"


class RunStatus(str, Enum):
    PASS = "pass"
    FAIL = "fail"


class Severity(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class Confidence(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class GateName(str, Enum):
    LINT = "lint"
    TYPECHECK = "typecheck"
    TEST = "test"


class ActionScope(str, Enum):
    SINGLE_FILE = "single_file"
    MULTI_FILE = "multi_file"
    CROSS_MODULE = "cross_module"


# --- Gate execution models ---


class GateResult(BaseModel):
    name: GateName
    status: GateStatus
    duration_ms: int = 0


class Finding(BaseModel):
    id: str
    gate: GateName
    severity: Severity
    summary: str
    files: list[str] = Field(default_factory=list)
    rule: str | None = None
    line: int | None = None
    column: int | None = None
    actual: int | str
    threshold: int | str
    status: str = "fail"
    raw: dict[str, Any] = Field(default_factory=dict)


class InferredHint(BaseModel):
    finding_id: str
    hint: str
    confidence: Confidence


# --- Artifact payloads ---


class FailuresPayload(BaseModel):
    version: str = "1.0.0"
    run_id: str
    mode: RunMode
    status: RunStatus
    timestamp: str
    repo: str | None = None
    branch: str | None = None
    changed_files: list[str] = Field(default_factory=list)
    gates: list[GateResult]
    findings: list[Finding]
    inferred_hints: list[InferredHint] = Field(default_factory=list)


class CommandTrace(BaseModel):
    command: str
    cwd: str
    started_at: str
    duration_ms: int
    exit_code: int
    timed_out: bool = False
    stdout: str = ""
    stderr: str = ""


class EnvironmentInfo(BaseModel):
    python_version: str
    platform: str
    virtualenv: str | None = None
    resolver: str | None = None
    installed_packages: dict[str, str] = Field(default_factory=dict)


class RunMetadata(BaseModel):
    run_id: str
    mode: RunMode
    started_at: str
    completed_at: str
    duration_ms: int
    config_source: str
    environment: EnvironmentInfo
    command_traces: list[CommandTrace]


# --- Agent brief models ---


class PriorityAction(BaseModel):
    finding_id: str
    action: str
    scope: ActionScope
    target_files: list[str] = Field(default_factory=list)
    rationale: str


class RetryPolicy(BaseModel):
    max_attempts: int
    max_patch_lines: int
    abort_on_no_improvement: int


class EscalationInfo(BaseModel):
    required: bool
    reason_code: str | None = None
    message: str | None = None


class AgentBrief(BaseModel):
    run_id: str
    mode: RunMode
    status: RunStatus
    summary: str
    priority_actions: list[PriorityAction]
    retry_policy: RetryPolicy
    escalation: EscalationInfo | None = None


# --- Repair models ---


class RepairAttempt(BaseModel):
    attempt: int
    patch_lines: int
    before_findings: int
    after_findings: int
    improved: bool
    worsened: bool
    status: str
    actions: list[dict[str, Any]] = Field(default_factory=list)


class RepairReport(BaseModel):
    status: str
    attempts: list[RepairAttempt] = Field(default_factory=list)


class Escalation(BaseModel):
    status: str = "escalated"
    reason_code: str
    message: str
    evidence: dict[str, Any] = Field(default_factory=dict)



from __future__ import annotations

from pathlib import Path

from pygate.config import load_config
from pygate.constants import AGENT_BRIEF_JSON_FILE, AGENT_BRIEF_MD_FILE, PYGATE_DIR, EscalationCode
from pygate.fs import ensure_dir, read_json, write_json, write_text
from pygate.models import (
    ActionScope,
    AgentBrief,
    EscalationInfo,
    FailuresPayload,
    PriorityAction,
    RetryPolicy,
    RunStatus,
)

_GATE_ACTIONS: dict[str, str] = {
    "lint": "Apply targeted ruff fixes and re-run lint deterministically.",
    "typecheck": "Resolve Pyright type errors for impacted files and re-run typecheck.",
    "test": "Fix failing tests and ensure pytest passes.",
}


def _scope_for_finding(finding_files: list[str]) -> ActionScope:
    if len(finding_files) == 1:
        return ActionScope.SINGLE_FILE
    if len(finding_files) <= 3:
        return ActionScope.MULTI_FILE
    return ActionScope.CROSS_MODULE


def execute_summarize(*, input_path: str, cwd: Path | None = None) -> dict:
    cwd = cwd or Path.cwd()
    pygate_dir = cwd / PYGATE_DIR
    ensure_dir(pygate_dir)

    config = load_config(cwd)
    policy = config.get("policy", {})

    failures_data = read_json(Path(input_path))
    failures = FailuresPayload(**failures_data)

    priority_actions = [
        PriorityAction(
            finding_id=f.id,
            action=_GATE_ACTIONS.get(f.gate.value, f"Address {f.gate.value} failure."),
            scope=_scope_for_finding(f.files),
            target_files=f.files,
            rationale=f"{f.gate.value} failed deterministically. Address this before any inferred optimizations.",
        )
        for f in failures.findings
    ]

    count = len(failures.findings)
    summary = (
        "All deterministic gates passed."
        if failures.status == RunStatus.PASS
        else f"{count} deterministic finding(s) require repair."
    )

    escalation = (
        EscalationInfo(required=False)
        if failures.status == RunStatus.PASS
        else EscalationInfo(
            required=True,
            reason_code=EscalationCode.UNRESOLVED_DETERMINISTIC_FAILURES,
            message="Escalate with evidence packet if bounded repair loop cannot clear deterministic failures.",
        )
    )

    brief = AgentBrief(
        run_id=failures.run_id,
        mode=failures.mode,
        status=failures.status,
        summary=summary,
        priority_actions=priority_actions,
        retry_policy=RetryPolicy(
            max_attempts=policy.get("max_attempts", 3),
            max_patch_lines=policy.get("max_patch_lines", 150),
            abort_on_no_improvement=policy.get("abort_on_no_improvement", 2),
        ),
        escalation=escalation,
    )

    brief_json_path = cwd / AGENT_BRIEF_JSON_FILE
    brief_md_path = cwd / AGENT_BRIEF_MD_FILE

    write_json(brief_json_path, brief.model_dump(mode="json"))
    write_text(brief_md_path, _generate_markdown(brief, failures))

    return {
        "brief_json_path": str(brief_json_path),
        "brief_md_path": str(brief_md_path),
        "status": brief.status.value,
    }


def _generate_markdown(brief: AgentBrief, failures: FailuresPayload) -> str:
    lines = [
        f"# PyGate Agent Brief — {brief.run_id}",
        "",
        f"**Mode:** {brief.mode.value}  ",
        f"**Status:** {brief.status.value}  ",
        f"**Summary:** {brief.summary}",
        "",
    ]

    if brief.priority_actions:
        lines.append("## Findings & Actions")
        lines.append("")
        for action in brief.priority_actions:
            lines.append(f"### `{action.finding_id}`")
            lines.append(f"- **Action:** {action.action}")
            lines.append(f"- **Scope:** {action.scope.value}")
            if action.target_files:
                lines.append(f"- **Files:** {', '.join(action.target_files)}")
            lines.append(f"- **Rationale:** {action.rationale}")
            lines.append("")

    lines.append("## Retry Policy")
    lines.append("")
    lines.append(f"- Max attempts: {brief.retry_policy.max_attempts}")
    lines.append(f"- Max patch lines: {brief.retry_policy.max_patch_lines}")
    lines.append(f"- Abort on no improvement: {brief.retry_policy.abort_on_no_improvement} consecutive attempts")
    lines.append("")

    if brief.escalation:
        lines.append("## Escalation")
        lines.append("")
        lines.append(f"- Required: {brief.escalation.required}")
        if brief.escalation.reason_code:
            lines.append(f"- Reason: {brief.escalation.reason_code}")
        if brief.escalation.message:
            lines.append(f"- Message: {brief.escalation.message}")
        lines.append("")

    return "\n".join(lines)
